import { proxyCustomElement, HTMLElement, h } from '@stencil/core/internal/client';

const sectionCss = ":host{display:block}.section__title{margin:0 0 var(--gap-sm);font-size:2rem;font-weight:500}.section__link{display:inline-block;color:rgb(var(--bd-violet-rgb));font-size:1.4rem}.section__link::after{content:\" ↗️\"}.section__code{margin:0;padding:var(--gap-sm);border:1px solid rgba(var(--bd-violet-rgb), .2);background-color:#f8f9fa;border-radius:.6rem;border-top-right-radius:0;border-top-left-radius:0;color:#333;font-size:1.4rem}.section__example{padding:var(--gap-sm) var(--gap-sm) var(--gap);border:1px solid rgba(var(--bd-violet-rgb), .2);border-bottom:none;border-radius:.6rem;border-bottom-left-radius:0;border-bottom-right-radius:0;background-color:rgba(var(--bd-violet-rgb), 0.02)}.section__code-title,.section__example-title{margin:0 0 var(--gap-sm);color:#777;font-weight:300;font-size:1.3rem;text-transform:uppercase;font-family:monospace}.section__example-content{margin:0 var(--gap-sm)}.section>*+*{margin-top:var(--gap-sm)}";

const BSWCSection = /*@__PURE__*/ proxyCustomElement(class BSWCSection extends HTMLElement {
  constructor() {
    super();
    this.__registerHost();
    this.__attachShadow();
    this.heading = undefined;
    this.slug = undefined;
  }
  formatHtml(content) {
    let string = content.replaceAll('<!---->', '');
    string = string.replaceAll(' class="hydrated"', '');
    let array = string.split('> <');
    array = array.map((element, index) => {
      const isFirst = index === 0;
      const isLast = index === array.length - 1;
      let html = '';
      if (!isFirst) {
        html += '<';
      }
      html += element;
      if (!isLast) {
        html += '>';
      }
      return html;
    });
    return array;
  }
  render() {
    const code = this.formatHtml(this.element.innerHTML);
    return (h("section", { class: "section" }, h("h4", { class: "section__title" }, this.heading), h("div", null, h("div", { class: "section__example" }, h("div", { class: "section__example-title" }, "Example"), h("div", { class: "section__example-content" }, h("slot", null))), h("pre", { class: "section__code" }, h("h5", { class: "section__code-title" }, "HTML"), code.map(tag => {
      return h("div", null, h("code", null, tag.includes('bs-row') ? tag : '  ' + tag));
    }))), this.slug ?
      h("a", { href: `https://getbootstrap.com/docs/5.3/layout/grid/#${this.slug}`, target: "_blank", class: "section__link" }, "More info about ", this.heading, " at Bootstrap website")
      : ''));
  }
  get element() { return this; }
  static get style() { return sectionCss; }
}, [1, "bswc-section", {
    "heading": [1],
    "slug": [1]
  }]);
function defineCustomElement$1() {
  if (typeof customElements === "undefined") {
    return;
  }
  const components = ["bswc-section"];
  components.forEach(tagName => { switch (tagName) {
    case "bswc-section":
      if (!customElements.get(tagName)) {
        customElements.define(tagName, BSWCSection);
      }
      break;
  } });
}

const BswcSection = BSWCSection;
const defineCustomElement = defineCustomElement$1;

export { BswcSection, defineCustomElement };

//# sourceMappingURL=bswc-section.js.map